#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2009 Etienne Robillard <robillard.etienne@gmail.com>
# All rights reserved.
"""
BCLO Controller API

BaseControllerLikeObjects (BCLOs) aims to provide a heterogeneous 
environment for extending Django web applications on top of the WSGI 
(PEP-333) spec.

""" 
# WebOb and Standard libraries
import sys, urllib
#from cStringIO import StringIO

from webob.exc import (
    HTTPInternalServerError, 
    HTTPServerError, 
    HTTPNotFound
    )
from notmm.datastore import LocalStore

#from _logging import StreamErrorHandler
#import louie

# Required third-party modules
try:
    from django.core.urlresolvers import get_resolver
except ImportError:
    def get_resolver(urlconf):
        # XXX defining a generic get_resolver method will solve all django related
        # problems we might had trying to access its core documentation with
        # pydoc. 
        raise NotImplementedError

def autoload(global_conf, **options):
    """Autoload a settings module. """
    
    if not isinstance(global_conf, str) and \
    hasattr(global_conf, 'module_name'):
        # for backward-compatibility; 
        mod = LocalStore(global_conf.module_name, **options)
    else:
        mod = LocalStore(global_conf, **options)
    
    return mod


__all__ = ('autoload', 'BaseControllerMixin', 'BaseController')

class BaseControllerMixin(object):
    """
    This ``BaseController`` class works by providing several methods  
    for routing requests to a matching Django view or callable function.
    
    """
    
    # Middlewares (for request/response fixes)
    request_middlewares = ()  # To be applied _before_ getresponse
    request_class = None
    response_middlewares = () # To be applied _after_ getresponse
    response_class = None
    
    # XXX Logging options (currently broken!)
    #logger = logging.Logger(__name__) 
    #errorLogger = logger 
    #errorHandler = StreamErrorHandler(sys.stderr)

    def __init__(self, 
        global_conf, #Global application settings 
        request_class=None,#XXX: DatabaseOpener screw this up 
        response_class=None, 
        urlconf=None, #XXX: should be request.urlconf...
        settings=None #Deprecated
        ):
        """Init options for BaseController-like apps"""
        
        #self.global_conf = global_conf
        #self.engine = getattr(global_conf, 'engine', None) # SA engine object
        
        self._request_class = request_class  
        self._response_class = response_class

        # Allow both syntax styles 
        for name in ['request_class', 'response_class']:
            if not getattr(self, name):
                # names supplied as keyword arguments have priority
                setattr(self, name, getattr(self, '%s' % str('_' + name)))
        
        #print "request_class = %r" % self.request_class

        self.mimetype = self.default_options.get('mimetype')
        self.charset  = self.default_options.get('charset')
        #self.strategy = self.default_options.get('strategy')
        #self.namespace = self.default_options.get('namespace')
        self.exc_info  = None   # For debugging purposes
        self.urlconf = urlconf  # This should point to ROOT_URLCONF
        self.headerlist = []    # List of headers as used in webob.Response
        self.response = None

        # initialize default logging options
        #self.logger.addHandler(self.errorHandler)

        self._environ = {}

        # fun experimenting with thread-local storage objects ;-)
        try:
            self.settings = autoload(global_conf, autoload=True)
        except (ImportError, AttributeError):    
            # work around: use django.conf.settings as
            # our last hope to make things work.
            from django.conf import settings
            setattr(self, 'settings', settings); del settings

        if self.urlconf is None:
            setattr(self, 'urlconf', getattr(self.settings, 'ROOT_URLCONF'))
    
    @classmethod
    def sethandle(cls, name, string_or_callable):
        """Adds custom response handlers to the BaseController class.
        
        This is for BCLO extensions that requires deeper control over the
        handling of generic HTTP responses.

        """

        # If string_or_callable is in the form of "foo.bar.quux",
        # the callable function should be the last part of the
        # string
        handler = None
        if isinstance(string_or_callable, str):
            if string_or_callable.find('.') != -1:
                bits = string_or_callable.rsplit('.', 1)
                m = __import__(bits[0], globals(), locals(), fromlist=[''])
                for component in bits[1:]:
                    if hasattr(m, component):
                        handler = getattr(m, component)
                        break
                    else:
                        continue
        elif callable(string_or_callable):
            handler = string_or_callable
        else:
            raise ValueError("Unexpected string_or_callable type: %r" % \
                type(string_or_callable))

        # Use staticmethod(func) here, because we want to support the
        # same positional arguments named by this callable
        if handler:
            handler = staticmethod(handler)
            setattr(cls, name, handler)
            # Register the callable object as a valid receiver
            #louie.connect(handler, sender=cls)

        return cls
    
    def environ_getter(self):
        """ Return the matching environ object from self._locals """
        return getattr(self, '_environ')

    environ = property(environ_getter)

    def getrequest(self):
        """ Return the request object for the current thread """
        return getattr(self, '_request')

    def setrequest(self, environ, **kwargs):
        """ Set options for the current WSGI request """
        # Update environ with self.environ
        self.environ.update(environ)
        request = self.request_class(getattr(self, 'environ'), **kwargs)
        
        # Options to add to the request obj
        request_quirks = (
            ('settings', self.settings),
            ('urlconf', self.urlconf),
            )

        # Adds "settings" and "urlconf" in the request object
        def _apply_request_fixes(request, options):
            for label, value in options:
                setattr(request, label, value)
            
            return request

        request = _apply_request_fixes(request, request_quirks)
        setattr(self, '_request', request)

        return None

    request = property(getrequest, setrequest)

    def getresponse(self, environ, start_response):
        """Return a WSGI response compatible with generic views"""
        
        # Remove reserved characters from request.path
        request = self.getrequest()

        # Encode the request.path with self.charset if its
        # a unicode type
        if isinstance(request.path, unicode):
            path_url = urllib.quote(request.path.encode(self.charset))
        else:
            path_url = urllib.quote(request.path)
        
        resolver = get_resolver(self.urlconf)
        args, kwargs = (), {}

        try:
            callback, args, kwargs = resolver.resolve(path_url)
            response = callback(request, *args, **kwargs)
            if response is not None:
                return response
        except (HTTPNotFound, HTTPServerError):
            # Store the sys.exc_info returned value
            # for later use.
            setattr(self, 'exc_info', sys.exc_info())
            
            # Match the appropriate response handler based on the
            # status_int value (XXX: currently webob specific)
            if hasattr(self.exc_info[1], 'status_int'):
                response_status = self.exc_info[1].status_int
            else:
                response_status = 500

            response_handler_name = 'handle' + str(response_status)
            # sethandle does this for us..
            if hasattr(self, response_handler_name):
                response_handler = getattr(self, response_handler_name)
                response = response_handler(request, **kwargs)
                return response
            raise HTTPInternalServerError("Missing required %r method!" % response_handler_name)
        except:
            # Other uncaught errors (Not reached)
            raise
        return None

    def application(self, environ, start_response, exc_info=None):
        """
        Override this to change the default request/response 
        handling. 

        This method is used for properly returning
        a WSGI response by calling ``getresponse``. The latter
        does the grunt work of routing the request to a proper 
        response callback function.

        
        """
        
        # Experiments shows that theses objects must be
        # set when the app is called.
        #setattr(self, 'start_response', start_response)
        #setattr(self, '_environ', environ.copy())
        #setattr(self, '_request', self.request_class(environ, \
        #    charset=self.charset))
        self.setrequest(environ, charset=self.charset)

        try:
            # set up additional request variables
            # self.setup_environ(environ)
            self.response = self.getresponse(environ, start_response)
            if self.settings.DEBUG:
                assert callable(self.response), 'response is not a function.'
                assert self.response is not None, 'response can\'t be None.'
                assert isinstance(self.response, self.response_class), \
                    type(self.response)
                assert hasattr(self.response, 'status')
                assert hasattr(self.response, 'headerlist')
                assert self.urlconf is not None, 'urlconf not set!'

            status, headerlist = self.response.status, self.response.headerlist

            # Add the exc_info to start_response here unless theres no
            # errors
            if self.exc_info == exc_info:
                start_response(status, headerlist, self.exc_info)
            else:
                start_response(status, headerlist)

            return self.response

        finally:
            if hasattr(self.response, 'close'):
                self.response.close() # Close client connection
                
    def __call__(self, environ, start_response, exc_info=None):
        return self.application(environ, start_response, exc_info)

class BaseController(BaseControllerMixin):

    default_options = {
        'mimetype' : 'text/plain',
        'charset'  : 'UTF-8',
       #'strategy' : 'threadlocal'
    }

