#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2009 Etienne Robillard <robillard.etienne@gmail.com>
# All rights reserved.

"""Thread-local Storage API"""

import os, copy

try:
    import threading
except ImportError:
    import dummy_thread as threading

from base import DataStore

__all__ = ('get_settings', 'LocalStore')

def get_settings(modname, **options):
    """ Returns a ready-to-use LocalStore instance. """
    
    settings_instance = LocalStore(modname, **options)
    return settings_instance

class LocalStore(DataStore, threading.local):
    """Wrapper to store local data in a single thread"""
   
    initialized = False
    
    def __init__(self, modname=None, autoload=False, debug=False):
        # This init code was borrowed from the 
        # _threading_local module.
        if self.initialized:
            raise SystemError('__init__ called too many times!')
        # automatically load the objects contained in
        # modname if autoload is True.
        if autoload is not False: self.loads(modname)

    def loads(self, modname):
        """ 
        Load objects from `modname` using the `__import__`
        builtin function.
        """
        head, tail = os.path.splitext(modname)
        module_name = tail.lstrip('.')
        # this must handle both absolute and relative imports
        try:
            m_obj = __import__(modname, {}, {}, fromlist=[module_name])
        except ImportError:
            raise
        else:
            if (m_obj is not None and not self.initialized):
                # Copy the precious settings into our thread-safe 
                # storing class.
                kwds = copy.copy(m_obj.__dict__)
                self.__dict__.update(**kwds)
                self.initialized = True
        
        return self

