#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2008, 2009 Etienne Robillard <robillard.etienne@gmail.com>
# All rights reserved.

"""Decorators functions returning a callable"""


from threading import Lock

try:
    from functools import wraps
except ImportError:
    from django.utils.functional import wraps

global _mtx
_mtx = Lock()

__all__ = ('with_session', 'with_schevo_database', 'permalink')

def with_session(engine=None):
    """
    Decorator function for attaching a `` Session`` instance
    as a keyword argument in ``request.environ``. 
    """
    def decorator(view_func):
        def _wrapper(request, *args, **kwargs):
            if engine is not None:
                from notmm.utils.sql.session import ScopedSession as scoped_session
                scoped_session.set_session(engine)
                request.environ['_scoped_session'] = getattr(scoped_session, 'session')
            return view_func(request, *args, **kwargs)
        return wraps(view_func)(_wrapper)
    return decorator

def with_schevo_database(app_conf, key='database'):
    """
    Decorator that adds a schevo database object reference
    in the ``request.environ`` dictionary.

    """
    def decorator(view_func, **kwargs):
        def _wrapper(request, *args, **kwargs):
            _mtx.acquire()
            try:
                from dbopener import DatabaseOpener
                request_middlewares = (DatabaseOpener,)
                for mw_method in request_middlewares:
                    callback = mw_method(request, app_conf)
                return view_func(callback, *args, **kwargs)
            finally:
                _mtx.release()
        return wraps(view_func)(_wrapper, **kwargs)
    return decorator

def permalink(view_func):
    """
    Decorator that calls urlresolvers.reverse() to return a URL using
    parameters returned by the decorated function "func".

    "view_func" should be a function that returns a tuple in one of the
    following formats:
        (viewname, viewargs)
        (viewname, viewargs, viewkwargs)
    """
    from django.core.urlresolvers import reverse
    def inner(*args, **kwargs):
        bits = view_func(*args, **kwargs)
        return reverse(bits[0], None, *bits[1:3])

    return inner
