#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""TestCollection API

A bunch of functions and classes to extend on the ``unittest.TestLoader`` 
class for loading custom test collections. 

Test collections are regular directories to search for test scripts ::

    [app:myapp]
    ...
    collections = fruits vegetable algae
    ...

"""

import unittest
import os
import re
import sys
import imp

__all__ = ('get_test_modules',
           'import_module',
           'is_test_script', 
           'test_script_re', 
           'TestCollection')

test_script_re = re.compile(r"test_\w+.py$")

def import_module(pathname):
    """Helper function to import a module or package with ``imp``.

    """

    head, modname = os.path.split(pathname)

    if not head in sys.path:
        sys.path.append(head)

    try:
        fh, pathname, description = imp.find_module(modname[:-3], \
            [head,])
        module = imp.load_module(modname[:-3], fh, pathname, \
            description)
        fh.close()
    except ImportError, e:
        msg = ImportError(e)
        raise ImportError('Failed to import %r ... (%s)' % (modname, msg))
    else:
        return module

    return None

def is_test_script(pathname):
    """ Determine if pathname is a test module and return True. """
    # For now, if its starting with prefix and
    # its ending with '.py' then return True.
    if isinstance(pathname, str) \
    and re.search(test_script_re, pathname):
        return True
    
    return False

def get_test_modules(collections, topdown=True):
    """ Return a list of usable ``unittest.TestCase`` objects. """

    found = []
    
    # make this a list if collections contains
    # a list of space-separated names
    if isinstance(collections, str):
        _collections = [ item.strip() for item in collections.split() ]
    else:
        # support list-style grape definition
        _collections = collections    

    for item in _collections:
        path = os.path.realpath(item)
        for dirname, names, files in sorted(os.walk(path, topdown=topdown)):
            basedir = os.path.relpath(dirname)
            if basedir in _collections:
                data = sorted([os.path.join(dirname, f) for f in files])
                found.extend([item for item in data if is_test_script(item)])
            else:
                continue
            break

    # Use a unique set to remove duplicates in a list:
    # http://stackoverflow.com/questions/479897/how-do-you-remove-duplicates-from-a-list-in-python
    return sorted(set(found))

class TestCollection(unittest.TestLoader):
    """ Helper class extending ``unittest.TestLoader``. 
    
    XXX: Put more documentation here.
    """
    
    default_label = "DEFAULT"

    def loadTestsFromNames(self, names):
        """ Return a ``unittest.TestSuite`` object.

        This method can be used to create a custom test suite from 
        a list of user-defined test directories ::  

        .. coding: python

            >>> loader = TestCollection()
            >>> suite = loader.loadTestsFromNames(['apple', 'strawberry', 'pear'])
        
        """

        suite = unittest.TestSuite()
        items = []
        for name in set(names):
            try:
                module = import_module(name)
            except ImportError, e:
                msg = ImportError(e)
                print 'failed to import %r .. %s' % (name, msg)
            else:
                if module is not None: 
                    items.extend(unittest.findTestCases(module))
             
        if len(items) >= 1:
            suite.addTests(items)
            #print "Found %d tests contained in %d modules." % (len(items), len(names))
            #print "Valid test cases: %d." % suite.countTestCases()
            
        return suite

    loadTestsFromList = loadTestsFromNames
    
